@@ -964,8 +964,15 @@ nodarg(Type *t, int fp)
 			fatal("nodarg: offset not computed for %T", t);
 		n->xoffset = t->width;
 		n->addable = 1;
+		n->orig = t->nname;
 		break;
 	}
+	
+	// Rewrite argument named _ to __,
+	// or else the assignment to _ will be
+	// discarded during code generation.
+	if(isblank(n))
+		n->sym = lookup("__");
 
 	switch(fp) {
 	default:
@@ -1152,6 +1159,7 @@ memname(Node *n, Type *t)
 	strcpy(namebuf, n->sym->name);
 	namebuf[0] = '.';	// keep optimizer from registerizing
 	n->sym = lookup(namebuf);
+	n->orig->sym = n->sym;
 }
 
 void
@@ -1828,6 +1836,7 @@ naddr(Node *n, Addr *a, int canemitcode)
 		a->offset = n->xoffset;
 		a->sym = n->left->sym;
 		a->type = D_PARAM;
+		a->node = n->left->orig;
 		break;
 
 	case ONAME:
@@ -1840,6 +1849,9 @@ naddr(Node *n, Addr *a, int canemitcode)
 		}
 		a->offset = n->xoffset;
 		a->sym = n->sym;
+		a->node = n->orig;
+		//if(a->node >= (Node*)&n)
+		//	fatal("stack node");
 		if(a->sym == S)
 			a->sym = lookup(".noname");
 		if(n->method) {
@@ -1857,8 +1869,6 @@ naddr(Node *n, Addr *a, int canemitcode)
 			break;
 		case PAUTO:
 			a->type = D_AUTO;
-			if (n->sym)
-				a->node = n->orig;
 			break;
 		case PPARAM:
 		case PPARAMOUT:
@@ -1881,6 +1891,7 @@ naddr(Node *n, Addr *a, int canemitcode)
 			a->dval = mpgetflt(n->val.u.fval);
 			break;
 		case CTINT:
+		case CTRUNE:
 			a->sym = S;
 			a->type = D_CONST;
 			a->offset = mpgetfix(n->val.u.xval);
